# -*- coding: utf-8 -*-

from psycopg2.extensions import cursor
from enum import Enum
from .information_schema import columns, primary_key


class InvalidColumn(Exception):
    pass


def table_parts(name: str) -> (str, str):
    """
    Returns a tuple with schema and table names
    :param name:
    :return:
    """
    if name and '.' in name:
        return name.split('.', 1)
    else:
        return 'public', name


def list2str(elements: list, sep: str= ', ', prepend: str='', append: str='', prepend_to_list: str='') -> str:
    """
    Prepend to all strings in the list
    :param elements:
    :param sep: separator
    :param prepend:
    :param append:
    :param prepend_to_list: prepend to the return string, if elements is not None or empty
    :return:
    """
    if elements is None or len(elements) == 0:
        return ''
    return prepend_to_list + sep.join([prepend+x+append for x in elements])


def column_alias(column: str,
                 remap_columns: dict = {},
                 prefix: str= None,
                 field_if_no_alias: bool = False,
                 prepend_as: bool = False) -> list:
    """

    :param table_alias:
    :param column:
    :param field_if_no_alias: if True, return the field if the alias doesn't exist. If False return an empty string
    :param prepend_as: prepend " AS " to the alias
    :return: empty string if there is no alias and (i.e = field name)
    """
    col_alias = ''
    if column in remap_columns:
        col_alias = remap_columns[column]
    elif prefix:
        col_alias = prefix + column
    elif field_if_no_alias:
        col_alias = column
    if prepend_as and col_alias:
        col_alias = ' AS {al}'.format(al=col_alias)
    return col_alias


def select_columns(pg_cur: cursor,
                   table_schema: str,
                   table_name: str,
                   table_alias: str=None,
                   table_type: str = 'table',
                   remove_pkey: bool=True,
                   skip_columns: list=[],
                   remap_columns: dict = {},
                   columns_on_top: list=[],
                   columns_at_end: list=[],
                   prefix: str= None,
                   indent: int=2) -> str:
    """

    :param pg_cur: the psycopg cursor
    :param table_schema: the schema
    :param table_name: the name of the table
    :param table_type: the type of table, i.e. view or table
    :param table_alias: if not specified, table is used
    :param remove_pkey: if True, the primary is removed from the list
    :param skip_columns: list of columns to be skipped
    :param remap_columns: dictionary to remap columns
    :param columns_on_top: bring the columns to the front of the list
    :param columns_at_end: bring the columns to the end of the list
    :param prefix: add a prefix to the columns (do not applied to remapped columns)
    :param indent: add an indent in front
    :return:
    """
    cols = sorted(columns(pg_cur,
                          table_schema=table_schema,
                          table_name=table_name,
                          table_type=table_type,
                          remove_pkey=remove_pkey,
                          skip_columns=skip_columns),
                  key=lambda col: __column_priority(col))

    # check arguments
    for param, dict_or_list in {'remap_columns': remap_columns,
                                'columns_on_top': columns_on_top,
                                'columns_at_end': columns_at_end}.items():
        for col in dict_or_list:
            if col not in cols:
                raise InvalidColumn('Invalid column in {param} paramater: "{tab}" has no column "{col}"'
                                    .format(param=param, tab=table_name, col=col))


    return ',\n'.join(['{indent}{table_alias}.{column}{col_alias}'
                      .format(indent=indent*' ',
                              table_alias=table_alias or table_name,
                              column=col,
                              col_alias=column_alias(col, remap_columns=remap_columns, prefix=prefix, prepend_as=True))
                       for col in cols])


def insert_command(pg_cur: cursor,
                   table_schema: str,
                   table_name: str,
                   table_alias: str=None,
                   table_type: str = 'table',
                   remove_pkey: bool=True,
                   skip_columns: list=[],
                   comment_skipped: bool = True,
                   remap_columns: dict = {},
                   insert_values: dict = {},
                   columns_on_top: list=[],
                   columns_at_end: list=[],
                   prefix: str= None,
                   indent: int=2) -> str:
    """

    :param pg_cur: the psycopg cursor
    :param table_schema: the schema
    :param table_name: the name of the table
    :param table_type: the type of table, i.e. view or table
    :param table_alias: if not specified, table is used
    :param remove_pkey: if True, the primary is removed from the list
    :param skip_columns: list of columns to be skipped
    :param comment_skipped: if True, skipped columns are written but commented, otherwise they are not written
    :param remap_columns: dictionary to remap columns
    :param insert_values: dictionary of expression to be used at insert
    :param columns_on_top: bring the columns to the front of the list
    :param columns_at_end: bring the columns to the end of the list
    :param prefix: add a prefix to the columns (do not applied to remapped columns)
    :param indent: add an indent in front
    :return:
    """
    # get columns
    cols = sorted(columns(pg_cur,
                          table_schema=table_schema,
                          table_name=table_name,
                          table_type=table_type,
                          remove_pkey=remove_pkey),
                  key=lambda col: __column_priority(col))

    # check arguments
    for param, dict_or_list in {'skip_columns': skip_columns,
                                'remap_columns': remap_columns,
                                'insert_values': insert_values,
                                'columns_on_top': columns_on_top,
                                'columns_at_end': columns_at_end}.items():
        for col in dict_or_list:
            if col not in cols:
                raise InvalidColumn('Invalid column in {param} paramater: "{tab}" has no column "{col}"'
                                    .format(param=param, tab=table_name, col=col))

    return """{indent}INSERT INTO {s}.{t} (
{indent}      {cols} 
{indent}) VALUES ( 
{indent}      {new_cols}
{indent});
""".format(indent=indent*' ',
           s=table_schema,
           t=table_name,
           cols='\n{indent}    '
                .format(indent=indent*' ')
                .join(['{skip}{comma}{col}'.format(indent=indent*' ',
                                                   skip='-- ' if col in skip_columns else '',
                                                   comma=', ' if i else '',
                                                   col=col)
                            for i, col in enumerate(cols) if (comment_skipped or col not in skip_columns)]),
           new_cols='\n{indent}    '
                    .format(indent=indent*' ')
                    .join(['{skip}{comma}{value}'
                          .format(skip='-- ' if col in skip_columns else '',
                                  comma=', ' if i else '',
                                  value=insert_values.get(col,
                                                          'NEW.{cal}'.format(cal=column_alias(col,
                                                                                              remap_columns=remap_columns,
                                                                                              prefix=prefix,
                                                                                              field_if_no_alias=True))))
                                for i, col in enumerate(cols) if (comment_skipped or col not in skip_columns)]))


def update_command(pg_cur: cursor,
                   table_schema: str,
                   table_name: str,
                   table_alias: str=None,
                   table_type: str = 'table',
                   update_pkey: bool = False,
                   skip_columns: list=[],
                   comment_skipped: bool = True,
                   remap_columns: dict = {},
                   update_values: dict = {},
                   columns_on_top: list=[],
                   columns_at_end: list=[],
                   prefix: str= None,
                   where_clause: str = None,
                   indent: int=2) -> str:
    """
    Creates an UPDATE command
    :param pg_cur: the psycopg cursor
    :param table_schema: the schema
    :param table_name: the name of the table
    :param table_type: the type of table, i.e. view or table
    :param update_pkey: if True, the primary key will also be updated
    :param table_alias: if not specified, table is used
    :param skip_columns: list of columns to be skipped
    :param comment_skipped: if True, skipped columns are written but commented, otherwise they are not written
    :param remap_columns: dictionary to remap columns
    :param update_values: dictionary of expression to be used at insert
    :param columns_on_top: bring the columns to the front of the list
    :param columns_at_end: bring the columns to the end of the list
    :param prefix: add a prefix to the columns (do not applied to remapped columns)
    :param where_clause: can be manually specified
    :param indent: add an indent in front
    :return: the SQL command
    """
    # get columns
    cols = sorted(columns(pg_cur,
                          table_schema=table_schema,
                          table_name=table_name,
                          table_type=table_type,
                          remove_pkey=not update_pkey and table_type != 'view'),
                  key=lambda col: __column_priority(col))

    pkey = '' if where_clause else primary_key(pg_cur, table_schema, table_name)

    # check arguments
    for param, dict_or_list in {'skip_columns': skip_columns,
                                'remap_columns': remap_columns,
                                'update_values': update_values,
                                'columns_on_top': columns_on_top,
                                'columns_at_end': columns_at_end}.items():
        for col in dict_or_list:
            if col not in cols and col != pkey:
                raise InvalidColumn('Invalid column in {param} paramater: "{tab}" has no column "{col}"'
                                    .format(param=param, tab=table_name, col=col))

    next_comma_printed = [False]
    def print_comma(next_comma_printed: bool, is_skipped: bool) -> bool:
        if is_skipped:
            return next_comma_printed[0]
        elif not next_comma_printed[0]:
            next_comma_printed[0] = True
            return False
        else:
            return True

    return """UPDATE {s}.{t} {a} SET
{indent}    {cols}
{indent}  WHERE {where_clause};"""\
        .format(indent=indent*' ',
                s=table_schema,
                t=table_name,
                a=table_alias,
                cols='\n{indent}    '
                     .format(indent=indent*' ')
                     .join(['{skip}{comma}{col} = {new_col}'
                                .format(indent=indent*' ',
                                        skip='-- ' if col in skip_columns else '',
                                        comma=', ' if print_comma(next_comma_printed, col in skip_columns) else '',
                                        col=col,
                                        new_col=update_values.get(col,
                                                                  'NEW.{cal}'.format(cal=column_alias(col,
                                                                                                      remap_columns=remap_columns,
                                                                                                      prefix=prefix,
                                                                                                      field_if_no_alias=True))))
                                for i, col in enumerate(cols) if (comment_skipped or col not in skip_columns)]),
                where_clause=where_clause or '{pkey} = {pkal}'.format(pkey=pkey,
                                                                      pkal=update_values.get(pkey,
                                                                                             'OLD.{cal}'.format(cal=column_alias(pkey,
                                                                                                                                 remap_columns=remap_columns,
                                                                                                                                 prefix=prefix,
                                                                                                                                 field_if_no_alias=True)))))



def update_columns(columns: list, sep:str=', ') -> str:
    return sep.join(["{c} = NEW.{c}".format(c=col) for col in columns])


def __column_priority(column: str, columns_on_top: list=[], columns_at_end: list=[]) -> int:
    if column in columns_on_top:
        return 0
    elif column in columns_at_end:
        return 2
    else:
        return 1


# -*- coding: utf-8 -*-

import os
import psycopg2
import psycopg2.extras

from enum import Enum

from pirogue.utils import table_parts, list2str, update_columns
from pirogue.information_schema import TableHasNoPrimaryKey, columns, reference_columns, primary_key, default_value


class JoinType(Enum):
    INNER = 'INNER'
    LEFT = 'LEFT'
    RIGHT = 'RIGHT'
    FULL = 'FULL'


class Join:
    """
    Creates a simple join view with associated triggers to edit.
    """

    def __init__(self, master_table: str, joined_table: str,
                 pg_service: str=None,
                 view_schema: str=None,
                 view_name: str=None,
                 join_type: JoinType=JoinType.LEFT):
        """
        Produces the SQL code of the join table and triggers
        :param pg_service:
        :param master_table:
        :param joined_table:
        :param view_schema: the schema where the view will written to
        :param view_name: the name of the created view, defaults to vw_{master_table}_{join_table}
        :param join_type: the type of join
        """

        if pg_service is None:
            pg_service = os.getenv('PGSERVICE')
        self.conn = psycopg2.connect("service={0}".format(pg_service))
        self.cursor = self.conn.cursor()

        (self.master_schema, self.master_table) = table_parts(master_table)
        (self.joined_schema, self.joined_table) = table_parts(joined_table)

        self.joined_type = join_type

        if view_schema is None:
            if self.master_schema != self.joined_schema:
                raise ValueError('Destination schema cannot be guessed if different on sources tables.')
            else:
                self.view_schema = self.master_schema
        else:
            self.view_schema = view_schema

        self.view_name = view_name or "vw_{ta}_{tb}".format(ta=self.master_table, tb=self.joined_table)

        self.master_cols = columns(self.cursor, self.master_schema, self.master_table)
        self.joined_cols = columns(self.cursor, self.joined_schema, self.joined_table)
        (self.ref_master_key, self.ref_joined_key) = reference_columns(self.cursor, self.master_schema, self.master_table, self.joined_schema, self.joined_table)
        try:
            self.master_pkey = primary_key(self.cursor, self.master_schema, self.master_table)
        except TableHasNoPrimaryKey:
            self.master_pkey = self.ref_master_key
        self.joined_pkey = primary_key(self.cursor, self.joined_schema, self.joined_table)
        self.joined_cols_wo_pkey = columns(self.cursor, self.joined_schema, self.joined_table, remove_pkey=True)
        self.master_cols_wo_pkey = list(self.master_cols)  # make a copy, otherwise keeps reference
        self.master_cols_wo_pkey.remove(self.master_pkey)

    def create(self) -> bool:
        """

        :return:
        """

        for sql in [self.__view(),
                    self.__insert_trigger(),
                    self.__update_trigger(),
                    self.__delete_trigger()]:
            try:
                self.cursor.execute(sql)
            except psycopg2.Error as e:
                print("*** Failing:\n{}\n***".format(sql))
                raise e
        self.conn.commit()
        self.conn.close()
        return True

    def __view(self) -> str:
        """
        Create the SQL code for the view
        :return: the SQL code
        """
        sql = """
CREATE OR REPLACE VIEW {ds}.{dt} AS SELECT
  {master_cols}{joined_cols}
  FROM {sm}.{tm} AS m
  {jt} JOIN {sj}.{tj} AS j ON j.{rjk} = m.{rmk};
""".format(ds=self.view_schema,
           dt=self.view_name,
           jt=self.joined_type.value,
           sm=self.master_schema,
           tm=self.master_table,
           rmk=self.ref_master_key,
           master_cols=list2str(self.master_cols, prepend='m.'),
           sj=self.joined_schema,
           tj=self.joined_table,
           rjk=self.ref_joined_key,
           joined_cols=list2str(self.joined_cols_wo_pkey, prepend='j.', prepend_to_list=', '))
        return sql

    def __insert_trigger(self) -> str:
        """

        :return:
        """
        sql = """
-- INSERT TRIGGER
CREATE OR REPLACE FUNCTION {ds}.ft_{dt}_insert() RETURNS trigger AS
$BODY$
BEGIN
INSERT INTO {sj}.{tj} ( {joined_cols} )
  VALUES (
    COALESCE( NEW.{rmk}, {jkp_def} ), {joined_new_cols} )
  RETURNING {jpk} INTO NEW.{rmk};
INSERT INTO {sm}.{tm} ( {master_cols} )
  VALUES (
    COALESCE( NEW.{mpk}, {mkp_def} ),  {master_new_cols} );
RETURN NEW;
END;
$BODY$
LANGUAGE plpgsql;

CREATE TRIGGER tr_{dt}_on_insert
  INSTEAD OF INSERT ON {ds}.{dt}
  FOR EACH ROW EXECUTE PROCEDURE {ds}.ft_{dt}_insert();
""".format(ds=self.view_schema,
           dt=self.view_name,
           sm=self.master_schema,
           tm=self.master_table,
           rmk=self.ref_master_key,
           mpk=self.master_pkey,
           master_cols=list2str(self.master_cols, prepend='\n    '),
           mkp_def=default_value(self.cursor, self.master_schema, self.master_table, self.master_pkey),
           master_new_cols=list2str(self.master_cols_wo_pkey, prepend='\n    NEW.', append=''),
           sj=self.joined_schema,
           tj=self.joined_table,
           joined_cols=list2str(self.joined_cols, prepend='\n    '),
           jpk=self.joined_pkey,
           jkp_def=default_value(self.cursor, self.joined_schema, self.joined_table, self.joined_pkey),
           joined_new_cols=list2str(self.joined_cols_wo_pkey, prepend='\n    NEW.', append=''))
        return sql

    def __update_trigger(self):
        sql = """
-- UPDATE TRIGGER
CREATE OR REPLACE FUNCTION {ds}.ft_{dt}_update() RETURNS trigger AS
$BODY$
BEGIN
  UPDATE {sa}.{ta}
    SET {master_up_cols}
    WHERE {apk} = OLD.{apk};
  UPDATE {sb}.{tb}
    SET {joined_up_cols}
    WHERE {bpk} = OLD.{rak};
RETURN NEW;
END;
$BODY$
LANGUAGE plpgsql;

CREATE TRIGGER tr_{dt}_on_update
  INSTEAD OF UPDATE ON {ds}.{dt}
  FOR EACH ROW EXECUTE PROCEDURE {ds}.ft_{dt}_update();
""".format(ds=self.view_schema,
           dt=self.view_name,
           sa=self.master_schema,
           ta=self.master_table,
           apk=self.master_pkey,
           master_up_cols=update_columns(self.master_cols_wo_pkey, sep='\n    , '),
           sb=self.joined_schema,
           tb=self.joined_table,
           bpk=self.joined_pkey,
           rak=self.ref_master_key,
           joined_up_cols=update_columns(self.joined_cols_wo_pkey, sep='\n    , '))
        return sql

    def __delete_trigger(self):
        sql = """
CREATE OR REPLACE FUNCTION {ds}.ft_{dt}_delete() RETURNS trigger AS
$BODY$
BEGIN
  DELETE FROM {sa}.{ta} WHERE {apk} = OLD.{apk};
  DELETE FROM {sb}.{tb} WHERE {bpk} = OLD.{rak};
RETURN NULL;
END;
$BODY$
LANGUAGE plpgsql;

CREATE TRIGGER tr_{dt}_on_delete
  INSTEAD OF DELETE ON {ds}.{dt}
  FOR EACH ROW EXECUTE PROCEDURE {ds}.ft_{dt}_delete();
""".format(ds=self.view_schema,
           dt=self.view_name,
           sa=self.master_schema,
           ta=self.master_table,
           apk=self.master_pkey,
           sb=self.joined_schema,
           tb=self.joined_table,
           bpk=self.joined_pkey,
           rak=self.ref_master_key)
        return sql



from setuptools import setup
import sys

python_min_version = (3, 6)

if sys.version_info < python_min_version:
    sys.exit('Pirogue requires at least Python version {vmaj}.{vmin}.\n'
             'You are currently running this installation with\n\n{curver}'.format(
        vmaj=python_min_version[0],
        vmin=python_min_version[1],
        curver=sys.version))

setup(
    name = 'pirogue',
    packages = [
        'pirogue'
    ],
    scripts = [
        'scripts/pirogue'
    ],
    version = '0.4',
    description = 'PostgreSQL view generator',
    author = 'Denis Rouzaud',
    author_email = 'denis.rouzaud@gmail.com',
    url = 'https://github.com/opengisch/pirogue',
    download_url = 'https://github.com/opengisch/pirogue/archive/0.4.tar.gz', # I'll explain this in a second
    keywords = ['postgres'],
    classifiers = [
        'Topic :: Database',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Information Technology',
        'Development Status :: 3 - Alpha'
    ],
    install_requires = [
        'psycopg2-binary>=2.7.3'
    ],
    python_requires=">={vmaj}.{vmin}".format(vmaj=python_min_version[0], vmin=python_min_version[1]),
)
